;
;	ClientSupport.asm
;

.386
.model flat, stdcall
option casemap:none

; ------------------------------------------------------------

include Client.inc

; ------------------------------------------------------------

.const

LOC_BUF_SIZE					equ 128

.data

ClientConn								db "Client Conn Mgmt", 0

wsaSockErr								db "Socket", 0
wsaSyncErr								db "WSAAsyncSelect", 0
connectErr								db "FD_CONNECT", 0
readRecvErr							db "FD_READ recv", 0
sendErr									db "send", 0
allocErr									db "Failure to allocate memory", 0

msgFromSenderFmt				db "Msg from %s: ", 0

blankStr									db " ", 0

lpQueueBuffer						DWORD 0
queueBufLen							DWORD 0
queueBytesSent						DWORD 0

.data?

locBuf										db LOC_BUF_SIZE dup (?)
recvBuffer								db BUFFER_SIZE dup (?)
sendBuffer								db BUFFER_SIZE dup (?)
senderName							db CLIENT_NAME_SIZE dup (?)

sockAddr									sockaddr_in <>

; ------------------------------------------------------------

.code

; ------------------------------------------------------------
;	InitSocket PROTO :HWND
;		Description: Initialize socket for the client function-
;			ality.
;		Return: Socket, else 0 indicating failure to initialize.
; ------------------------------------------------------------

InitSocket PROC  hwnd:HWND

LOCAL retVal :DWORD
LOCAL sock :DWORD

	mov retVal, 0 ; error val

		; create the listening socket
	invoke socket, AF_INET, SOCK_STREAM, IPPROTO_TCP 

	.if eax == INVALID_SOCKET

		invoke WSAputError,addr wsaSockErr

	.else

		mov sock, eax 

			; add asynchronous behavior messages
		invoke WSAAsyncSelect, eax, hwnd, WM_SOCKET, 
						FD_READ+\
						FD_CLOSE+\
						FD_CONNECT + \
						FD_WRITE

		.if eax == SOCKET_ERROR

			invoke WSAputError,addr wsaSyncErr
			mov sock, 0

		.endif

	.endif

	mov eax, sock
	ret ; return socket, else error is eax == 0

InitSocket endp

; ------------------------------------------------------------
; QueueConnection PROTO :DWORD, :DWORD
;		Description: Sets up the connection and sends message
;			with FD_CONNECT to WndProc to execute.
;		Return: Nothing.
; ------------------------------------------------------------

QueueConnection PROC sock:DWORD, lpHostAddr:DWORD

	invoke memfill, addr sockAddr, SIZEOF sockaddr_in, 0

	mov ax, SERVER_PORT
	invoke htons, ax
	mov sockAddr.sin_port, ax

	mov sockAddr.sin_family, PF_INET

	invoke inet_addr, lpHostAddr
	
	.if eax == INADDR_NONE
	
		invoke WSAputError, chr$( "Invalid host address" )

	.else

		mov sockAddr.sin_addr, eax

		invoke GetActiveWindow ; sends back to WndProc
		mov edx, FD_CONNECT ; high word == 0
		invoke PostMessage, eax, WM_SOCKET, sock, edx
	
	.endif

	xor eax, eax ; no return, the PostMessage goes or doesn't
	ret 

QueueConnection endp

; ------------------------------------------------------------
; ConnectToServer	PROTO :SOCKET
;		Description: Execute connect to link to server.
;		Return: Success = 1, failure = 0.
; ------------------------------------------------------------

ConnectToServer PROC sock:SOCKET

LOCAL retVal :DWORD

	mov retVal, 0

			; <<<<<<<<<<<< connect --> the main function >>>>>>>>>>>>
	invoke connect, sock, addr sockAddr, SIZEOF sockaddr_in

	.if eax == SOCKET_ERROR

			invoke WSAGetLastError

			.if eax != WSAEWOULDBLOCK

				invoke WSAputError, addr connectErr

; **** NOTE **** **** NOTE **** **** NOTE **** **** NOTE **** 
;		I believe that the WSAEWOULDBLOCK occurs automatically,
;			at least in non-blocking mode, since the server's
;			connection acceptance hasn't occurred immediately;
;			while debugging, I saw the connection show up on the
;			server anyway -- perhaps later than this connection
;			attempt recognized; I will test to see if this error
;			can be safely ignored. [Appeared to be the case.]
; **** NOTE **** **** NOTE **** **** NOTE **** **** NOTE **** 

			.else ; try again 

;				invoke GetActiveWindow ; sends back to WndProc
;				mov edx, FD_CONNECT ; high word == 0
;				invoke PostMessage, eax, WM_SOCKET, sock, edx

				mov retVal, 1 ; test

			.endif

	.else
	
		mov retVal, 1

	.endif
	
	mov eax, retVal
	ret ; return failure eax == 0, success = 1

ConnectToServer endp

; ------------------------------------------------------------
; QueueLoginName PROTO :DWORD, :DWORD
;		Description: Format message to the Server module that
;			tells it the client user's login name; then sends a
;			PostMessage with FD_WRITE for client's WndProc to execute
;		Return: Nothing.
; ------------------------------------------------------------

QueueLoginName PROC sock:DWORD, lpLoginName:DWORD

	invoke szLen, lpLoginName
	add eax, 1 ; for command char
	mov queueBufLen, eax
	
	invoke Alloc, eax
	
	.if eax == NULL
	
			invoke WSAputError, addr allocErr

	.else

		mov lpQueueBuffer, eax
		mov ecx, eax

		mov BYTE PTR [ecx], cmdHello

		add ecx, 1 ; bump past cmd
		invoke szCopy, lpLoginName, ecx

		mov queueBytesSent, 0

		invoke GetActiveWindow ; sends back to WndProc
		mov edx, FD_WRITE ; high word == 0
		invoke PostMessage, eax, WM_SOCKET, sock, edx

	.endif
	
	xor eax, eax
	ret

QueueLoginName endp

; ------------------------------------------------------------
; QueueMessage PROTO :DWORD, :DWORD, :DWORD
;		Description: Format a message to the Server that tells
;			it the client user wants a message forwarded to another
;			user on another client machine; performs a PostMessage 
;			with FD_WRITE for client's WndProc to execute.
;		Note: format of message sent to server is:
;				'UrecipnameXtextofmessage' 
;			where 'U' is the upload command, recipname is the 
;			text of the recipient's name, the 'X' is a blank 
;			character, and textofmessage is the text of the message.
;		Return: Nothing.
; ------------------------------------------------------------

QueueMessage PROC sock:DWORD, lpRecipient:DWORD, lpMessage:DWORD

LOCAL recipLen :DWORD

	invoke szLen, lpRecipient
	mov ecx, eax
	mov recipLen, eax
	
	invoke szLen, lpMessage
	add ecx, eax
	
	add ecx, 2 ; added for command char and name space delimiter
	mov queueBufLen, ecx ; total bytes eventually sent
	
	invoke Alloc, ecx
	
	.if eax == NULL
	
			invoke WSAputError, addr allocErr

	.else

		mov lpQueueBuffer, eax

		invoke memfill, lpQueueBuffer, queueBufLen, 0

		mov ecx, lpQueueBuffer
		mov BYTE PTR [ecx], cmdUpload ; put 'U' in first byte

		add ecx, 1 ; bump past cmd in buffer
		invoke szCopy, lpRecipient, ecx

		mov eax, recipLen
		add eax, 1 ; for cmd
		mov ecx, lpQueueBuffer
		add ecx, eax ; now pointing into buffer just after recip name
		mov BYTE PTR [ecx], 32 ; append a blank char
		
		add ecx, 1 ; bump past blank
		invoke szCopy, lpMessage, ecx ; now everything is loaded

		mov queueBytesSent, 0

		invoke GetActiveWindow ; sends back to WndProc
		mov edx, FD_WRITE ; high word == 0
		invoke PostMessage, eax, WM_SOCKET, sock, edx

	.endif
	
	xor eax, eax
	ret

QueueMessage endp

; ------------------------------------------------------------
; QueueGoodbye PROTO :DWORD
;		Description: Format a message to the Server that tells
;			it that the client is logging out.
;		Return: Nothing.
; ------------------------------------------------------------

QueueGoodbye PROC sock:DWORD

	invoke Alloc, 1
	
	.if eax == NULL
	
			invoke WSAputError, addr allocErr

	.else

		mov lpQueueBuffer, eax
		mov BYTE PTR [eax], cmdGoodbye
		mov queueBufLen, 1
		mov queueBytesSent, 0

		invoke GetActiveWindow ; sends back to WndProc
		mov edx, FD_WRITE ; high word == 0
		invoke PostMessage, eax, WM_SOCKET, sock, edx

	.endif
	
	xor eax, eax
	ret

QueueGoodbye endp

; ------------------------------------------------------------
; QueueWhoList PROTO :DWORD
;		Description: Format a message to the server to
;			request a list of all users logged on to server.
;		Return: Nothing.
; ------------------------------------------------------------

QueueWhoList PROC sock:DWORD

	invoke Alloc, 1
	
	.if eax == NULL
	
			invoke WSAputError, addr allocErr

	.else

		mov lpQueueBuffer, eax
		mov BYTE PTR [eax], cmdWho
		mov queueBufLen, 1
		mov queueBytesSent, 0

		invoke GetActiveWindow ; sends back to WndProc
		mov edx, FD_WRITE ; high word == 0
		invoke PostMessage, eax, WM_SOCKET, sock, edx

	.endif
	
	xor eax, eax
	ret

QueueWhoList endp

; ------------------------------------------------------------
;	RecvData PROTO :LPSOCKETCONN
;		Description: Processes incoming data. Parses incoming 
;			buffer for server responses (in the first byte).
;			There are 6 server responses:
;				'K' == Server connection established, login expected
; 			'L' == Server acknowledged login complete
;				'T' == Server response after log out (termination)
;				'E' == Error on last communication with server
;				'C' == Server reached end of upload of data
; 			'I' == Requested info download complete
;				'D' == Server is downloading data
;		Note: every command send to server will receive feedback
;		Return: Nothing.
; ------------------------------------------------------------

RecvData PROC sock:SOCKET

LOCAL bytesRecvd :DWORD
LOCAL stopLooping :DWORD
LOCAL downLoadTrans :DWORD
LOCAL lenSenderName :DWORD

	mov stopLooping, 0
	mov downLoadTrans, 0
	
	invoke memfill, addr senderName, CLIENT_NAME_SIZE, 0
	
	.repeat ; until stopLooping == 1

			; always re-initialize the receive buffer as all of
			;		its previous contents have been processed, if only
			;		temporarily, e.g., into a queue
		invoke memfill, addr recvBuffer, BUFFER_SIZE, 0

; <<<<<<<<<<<< RECV --> the main function >>>>>>>>>>>>

		invoke recv, sock, addr recvBuffer, BUFFER_SIZE, 0 ; 

; <<<<<<<<<<<< RECV --> the main function >>>>>>>>>>>>

		.if eax == SOCKET_ERROR

			invoke WSAGetLastError

			.if eax != WSAEWOULDBLOCK

				invoke WSAputError, addr readRecvErr
				
			.else ; was a blocking situation -- so no more data

					invoke PostToLog, chr$( "Msg end" )
	
			.endif

			mov stopLooping, 1 ; stop receiving upload for any error

		.else ; not a SOCKET_ERROR

			.if eax == 0 ; connection was closed

				mov stopLooping, 1 

					; documentation indicates that a recv 0 bytes len
					;		means that the sender has -- in the interim --
					;		committed a graceful close; if that is the
					;		case, the close signal has not yet arrived; 
					;		should it be signalled to the main message loop 
					;		from here as follows?

				invoke GetActiveWindow
				mov edx, FD_CLOSE ; high word == 0
				invoke PostMessage, eax, WM_SOCKET, sock, edx

			.else

				mov bytesRecvd, eax

					; It is possible that a send command was received in
					;		a previous iteration through this loop and that
					;		subsequent reads may occur; this data will be part
					;		of that stream and therefore will not carry a command;
					;		it merely needs to be queued up and forwarded to
					;		the recipient to follow the previous queued batch(es).

;
;parse the input to get the command
;
	
				mov edx, OFFSET recvBuffer
				mov al, BYTE PTR [edx]  ;command character is in 1st byte
				add edx, 1
	
				.if al == svrInfo  ; <<<<<<<<<<< 'I' command

						; this only happens after a 'who' request is filled
					invoke wsprintf, addr locBuf, chr$("Server info: %s" ), edx
					invoke PostToLog, addr locBuf

					mov stopLooping, 1

				.elseif al == srvError  ; <<<<<<<<<<< 'E' command

					invoke wsprintf, addr locBuf, chr$("Server error: %s" ), edx
					invoke PostToLog, addr locBuf
					mov stopLooping, 1

				.elseif al == svrCircuit  ; <<<<<<<<<<< 'K' command

						; occurs after connection is 'accept'ed
					invoke wsprintf, addr locBuf, chr$("Server connected: %s" ), edx
					invoke PostToLog, addr locBuf

					invoke GetActiveWindow
					invoke PostMessage, eax, WM_MENU_UPDATE, svrCircuit, 0

					mov stopLooping, 1

				.elseif al == srvLogCompl  ; <<<<<<<<<<< 'L' command

						; occurs after user logs in
					invoke wsprintf, addr locBuf, chr$("Login status: %s" ), edx
					invoke PostToLog, addr locBuf

					invoke GetActiveWindow
					invoke PostMessage, eax, WM_MENU_UPDATE, srvLogCompl, 0

					mov stopLooping, 1

				.elseif al == srvUploadCompl  ; <<<<<<<<<<< 'C' command

					;invoke wsprintf, addr locBuf, chr$("Upload status: %s" ), edx
					;invoke PostToLog, addr locBuf
					invoke PostToLog, edx

					mov stopLooping, 1

				.elseif al == srvTerminated ; <<<<<<<<<<< 'T' command

						;
						; Client does not shut down socket until
						;		it receives acknowledgment from the server
						;

					invoke wsprintf, addr locBuf, chr$( "Connection status: %s" ), edx
					invoke PostToLog, addr locBuf

					invoke GetActiveWindow
					invoke PostMessage, eax, WM_MENU_UPDATE, srvTerminated, 0

					mov stopLooping, 1

				.elseif al == srvDownload ; <<<<<<<<<<< 'D' command (msg from another client)

					push edx ; points to 1 byte past command

					invoke MemCopy, edx, addr senderName, CLIENT_NAME_SIZE
					invoke szTrim, addr senderName

					invoke InString, 1, addr senderName, addr blankStr

					.if eax > 0

						sub eax, 1
						mov lenSenderName, eax

						lea ecx, senderName
						add ecx, eax
						mov BYTE PTR [ecx], 0 ; put zero terminator

					.endif

					invoke wsprintf, addr locBuf, addr msgFromSenderFmt, \
													addr senderName

					invoke PostToLog, addr locBuf
					
					pop edx
					add edx, lenSenderName
					add edx, 1 ; for blank
					invoke PostToLog, edx

					mov downLoadTrans, 1

				.else ; recv buffer arrived w/o a command ???

					.if downLoadTrans == 1

						invoke PostToLog, recvBuffer

					.endif

				.endif

			.endif

		.endif

	.until stopLooping == 1

	xor eax,eax ;nothing to return
	ret
	
RecvData endp	

; ------------------------------------------------------------
;	SendData PROTO :SOCKET
;		Description: (Receives no input as it merely has to work
;			its way through the linked list queue of SENDPACKETs 
;			and forward those to their respective recipients.) Works 
;			in a loop as it dequeues the items in the queue and
;			forwards them; stops the loop when either the queue is
;			emptied or an attempt to send results in any error;
;			most common error will be WSAEWOULDBLOCK.
;		Return: Nothing.
; ------------------------------------------------------------

SendData PROC sock:SOCKET

LOCAL stopLooping :DWORD
LOCAL bufferSize :DWORD
LOCAL lostConnection :DWORD

	.if queueBufLen == 0
	
		xor eax, eax
		ret
		
	.endif
	
	mov stopLooping, 0
	mov lostConnection, 0

	.repeat ; until stopLooping == 1
	
		mov eax, queueBufLen
		sub eax, queueBytesSent
		
		.if eax == 0 ; buffer send complete, so stop sending

			mov stopLooping, 1

		.else

			mov edx, lpQueueBuffer
			add edx, queueBytesSent ; update offset into source buffer
			
			mov bufferSize, BUFFER_SIZE ; size of TCP output buffer

			.if eax > bufferSize ; note: comparing size incl. command len

				mov eax, bufferSize	; max size of output buffer

			.endif

			push eax
			invoke MemCopy, edx, addr sendBuffer, eax ; copy into TCP output buffer

			pop eax 

; <<<<<<<<<<<< SEND --> the main function >>>>>>>>>>>>

			invoke send, sock, addr sendBuffer, eax, 0 

; <<<<<<<<<<<< SEND --> the main function >>>>>>>>>>>>

			.if eax == SOCKET_ERROR

				invoke WSAGetLastError

				.if eax == WSAEWOULDBLOCK

						; not done, so set up another go-around for send
					invoke GetActiveWindow
					mov edx, FD_WRITE ; high word == 0
					invoke PostMessage, eax, WM_SOCKET, sock, edx
					invoke WSAputError, addr sendErr

				.else

						; assume that the recipient connection has been lost
						;		so we will need to remove its data; sending 
						;		WndProc an async message for FD_CLOSE should
						;		get a graceful close

					invoke GetActiveWindow
					mov edx, FD_CLOSE ; high word == 0
					invoke PostMessage, eax, WM_SOCKET, sock, edx
					mov lostConnection, 1

				.endif

				mov stopLooping, 1 ; stop for all error conditions

			.else ; at least something was sent

				add queueBytesSent, eax ; count bytes actually sent

			.endif

		.endif

	.until stopLooping == 1

	mov eax, queueBufLen
	sub eax, queueBytesSent ; eax contains bytes remaining

	.if eax == 0 || lostConnection == 1

		invoke Free, lpQueueBuffer ; release data buffer, not needed anymore
		mov lpQueueBuffer, NULL
		mov queueBufLen, 0
		mov  queueBytesSent, 0
	
	.endif

	xor eax,eax
	ret
	
SendData endp	

; ------------------------------------------------------------
; ClearRecvBufs	PROTO :SOCKET
;		Description: Execute recv on socket until all buffers have
;			been cleared; documentation indicates this is needed
;			even if no FD_READ was received -- to be on safe side?
;		Return: Nothing.
; ------------------------------------------------------------

ClearRecvBufs PROC sock:SOCKET 

LOCAL stopLoop :DWORD

	mov stopLoop, 0

	.repeat ; .until stopLoop == 1

		invoke recv, sock, addr recvBuffer, BUFFER_SIZE, 0 ; 

		.if eax == SOCKET_ERROR || eax == 0

			mov stopLoop, 1

		.endif

	.until stopLoop == 1

	xor eax, eax
	ret

ClearRecvBufs endp

; ------------------------------------------------------------
; ClearSendBufs	PROTO
;		Description: Verify that the sendbuffer has been freed.
;		Return: Nothing.
; ------------------------------------------------------------

ClearSendBufs PROC 

	.if lpQueueBuffer != NULL

		invoke Free, lpQueueBuffer ; release data buffer, not needed anymore

	.endif

	xor eax, eax
	ret

ClearSendBufs endp

; ------------------------------------------------------------

end