;
;	Server.asm
;
; -------------------------------------------------------------------------
;		Server.asm provides a Winsock-based relay tool for transferring
;			messages between client modules. It uses the asynchronous
;			model for Winsock processing, that is, it never blocks when
;			making a Winsock call. The method for bypassing blocking is
;			to initialize a new socket with the WSAAsyncSelect function,
;			with specifications for which processes should be communicated
;			as messages to the WndProc message loop. The main display of
;			both the server and client modules is an edit window which
;			is read-only. The edit window serves as a log of activities
;			and the user may save its contents to a file.
;		The server is entirely passive, i.e., it has no menu options
;			and it only relays messages received from clients. It
;			maintains a list of logged in clients. When a client sends
;			(i.e., uploads) a message to another client (no limit on
;			number that may be logged in), the server looks up the client
;			recipient name and forwards message to the recipient. A user
;			may log out and then log back in; however, any queued messages
;			will be lost. A client may send a message back to himself.
;		The server allows a user to save the contents of the editor
;			log (main display) to an output text file.
;		This software has been reasonably-well, but not exhaustively,
;			tested. It is provided here as a demonstration of asynchronous
;			Winsock processes between client and server operating on a 
;			simple relay model. Obviously, use it at your own risk.
;		All comments and suggestions will be appreciated.
; -------------------------------------------------------------------------

.386
.model flat, stdcall
option casemap:none

; -------------------------------------------------------------------------

include Server.inc ; loads all other include files

; -------------------------------------------------------------------------

WinMain PROTO :DWORD, :DWORD, :DWORD, :DWORD

; -------------------------------------------------------------------------

.const

IDM_SAVE						equ 2
IDM_EXIT							equ 3

EditID									equ 1

.data

ClassName							db "Win32ASMEditClass", 0
AppName							db "WSOCK Relay Server", 0
EditClass								db "edit", 0
MenuName						db "ServerMenu", 0
FilterString							db "Log Files", 0, "*.log", 0, 0
											db "All Files", 0, "*.*", 0

buffer									db MAXSIZE dup(0)
recvBuffer							db BUFFER_SIZE dup (0)

cleanErr								db "Cleanup", 0
closeErr								db "FD_CLOSE", 0
success								db "Successful so far!", 0
okay									db "Ok", 0
CrLf									db 13, 10, 0
CloseReq							db "Socket close request", 0
ConnectNow						db "Connection can go through now", 0

format									db "wHighVersion = %x", 0
highVer								db "Highest version at WSAStartup = %x !", 0
AsyncEventFmt					db "wsaAsyncEventErr:  %d"

lpCurrentSocketConn			LPSOCKETCONN 0

listenSocketConn				SOCKETCONN <>
lpListenSocketConn			LPSOCKETCONN listenSocketConn
sizeSocketConn					equ SIZEOF SOCKETCONN 
ofn										OPENFILENAME <>

.data?

currentSocket 						SOCKET		?
socketListen						SOCKET		?

hInstance								HINSTANCE	?
CommandLine					LPSTR			?
hwndEdit							HWND			?
hFile									HANDLE		?
hMemory							HANDLE		?
pMemory							DWORD		?
SizeReadWrite					DWORD		?

; -------------------------------------------------------------------------

.code

start:
	invoke GetModuleHandle, NULL
	mov hInstance, eax

	invoke GetCommandLine
	mov CommandLine, eax

	invoke WinMain, hInstance, NULL, CommandLine, SW_SHOWDEFAULT
	invoke ExitProcess, eax

; -------------------------------------------------------------------------

WinMain proc hInst:HINSTANCE, hPrevInst:HINSTANCE, CmdLine:LPSTR, CmdShow:DWORD

LOCAL wc:WNDCLASSEX
LOCAL msg:MSG
LOCAL hwnd:HWND

	mov	wc.cbSize, SIZEOF WNDCLASSEX
	mov	wc.style, CS_HREDRAW or CS_VREDRAW
	mov	wc.lpfnWndProc, OFFSET WndProc
	mov	wc.cbClsExtra, NULL
	mov	wc.cbWndExtra, NULL
	push	hInst
	pop	wc.hInstance
	mov	wc.hbrBackground, COLOR_WINDOW+1
	mov	wc.lpszMenuName, OFFSET MenuName
	mov	wc.lpszClassName, OFFSET ClassName

	invoke LoadIcon, NULL, IDI_APPLICATION
	mov wc.hIcon, eax
	mov wc.hIconSm, eax

	invoke LoadCursor, NULL, IDC_ARROW
	mov wc.hCursor, eax

	invoke RegisterClassEx, addr wc

	INVOKE CreateWindowEx, WS_EX_CLIENTEDGE, ADDR ClassName, ADDR AppName, \
			WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, \
			CW_USEDEFAULT, 400, 500, NULL, NULL, \
			hInst, NULL
	mov	hwnd, eax
	
	INVOKE ShowWindow, hwnd, SW_SHOWNORMAL
	INVOKE UpdateWindow, hwnd

	.WHILE TRUE
	
		INVOKE GetMessage, ADDR msg, NULL, 0, 0
		.BREAK .IF (!eax)

		INVOKE TranslateMessage, ADDR msg
		INVOKE DispatchMessage, ADDR msg
		
	.ENDW
		
	mov	eax, msg.wParam
	ret

WinMain endp

; -------------------------------------------------------------------------

WndProc proc uses ebx hWnd:HWND, uMsg:UINT, wParam:WPARAM, lParam:LPARAM

LOCAL wsaAsyncError :WORD
LOCAL wsaAsyncEvent :WORD

	.IF uMsg == WM_CREATE ; 

		INVOKE CreateWindowEx, NULL, ADDR EditClass, NULL, \
				WS_VISIBLE or WS_CHILD or ES_LEFT or ES_MULTILINE or \
				ES_AUTOHSCROLL or ES_AUTOVSCROLL or ES_READONLY, 0, \
				0, 0, 0, hWnd, EditID, hInstance, NULL ; ES_READONLY makes background gray
		mov hwndEdit, eax

		mov ofn.lStructSize, SIZEOF ofn
		push hWnd
		pop ofn.hWndOwner
		push hInstance
		pop ofn.hInstance
		mov ofn.lpstrFilter, OFFSET FilterString
		mov ofn.lpstrFile, OFFSET buffer
		mov ofn.nMaxFile, MAXSIZE

		invoke InitListen, hWnd, lpListenSocketConn
		.if eax == 0

			invoke ErrorMsg, chr$ ( "Failed initListen" ), addr AppName

			invoke shutdown, listenSocketConn.hSock, STOP_SENDS
			invoke closesocket, listenSocketConn.hSock
			invoke PostQuitMessage, WM_DESTROY

		.else

			invoke SetFocus, hwndEdit

		.endif

	.ELSEIF uMsg == WM_SIZE

		mov eax, lParam
		mov edx, eax
		shr edx, 16
		and eax, 0ffffh
		invoke MoveWindow, hwndEdit, 0, 0, eax, edx, TRUE

	.ELSEIF uMsg == WM_DESTROY

		invoke ClearSendQueue, NULL
		invoke GracefulDisconAll

		invoke shutdown, socketListen, STOP_SENDS
		invoke closesocket, socketListen
		
		invoke WSACleanup

		invoke PostQuitMessage, NULL

	.ELSEIF uMsg == WM_COMMAND
	
		mov eax, wParam

		.if lParam == 0 ; menu commands received

			.if ax == IDM_SAVE
			
				mov ofn.Flags, OFN_LONGNAMES or \
								OFN_EXPLORER or OFN_HIDEREADONLY

				invoke GetSaveFileName, ADDR ofn

				.if eax == TRUE

					invoke CreateFile, ADDR buffer, \
										GENERIC_READ or GENERIC_WRITE , \
										FILE_SHARE_READ or FILE_SHARE_WRITE, \
										NULL, CREATE_NEW, FILE_ATTRIBUTE_ARCHIVE, \
										NULL
					mov hFile, eax

					invoke GlobalAlloc, GMEM_MOVEABLE or GMEM_ZEROINIT, MEMSIZE
					mov hMemory, eax

					invoke GlobalLock, hMemory
					mov pMemory, eax

									; now get the text back from the edit control, 
									;	presumably into a buffer so it can be written
									;	to a file in a straightforward fashion.
					invoke SendMessage, hwndEdit, WM_GETTEXT, MEMSIZE-1, pMemory

									; okay, we can write to the file now
					invoke WriteFile, hFile, pMemory, eax, ADDR SizeReadWrite, NULL

									;
									; release everything now
									;
					invoke CloseHandle, hFile
					invoke GlobalUnlock, pMemory
					invoke GlobalFree, hMemory
				.endif

				invoke SetFocus, hwndEdit

			.else
				invoke DestroyWindow, hWnd
			.endif
		.endif

; <<<<<<<<<<<<<<<< >>>>>>>>>>>>>>>>>
;
;	The following block of options are a product of the 
;		asynchronous messages generated by the Winsock 
;		DLL: events identified by the WSAAsyncSelect function; 
;		the message I chose is referenced by the WM_SOCKET uMsg 
;		value.
;
	.elseif uMsg == WM_SOCKET

		push wParam
		pop currentSocket

		mov eax, lParam
		mov wsaAsyncEvent, ax

		shr eax, 16 ; async error code in the high word of lParam
		mov wsaAsyncError, ax

		.if wsaAsyncError == NULL

			invoke ConnFind, currentSocket
			mov lpCurrentSocketConn, eax

		.else

			invoke wsprintf, addr buffer, AsyncEventFmt, wsaAsyncEvent
			invoke WSAputError,addr buffer
			mov wsaAsyncEvent, NULL ; so no processing occurs
	
		.endif
	
;
; FD_CLOSE occurs when the socket is being closed by the
;		remote connection
;
		.if wsaAsyncEvent == FD_CLOSE 

			.if lpCurrentSocketConn != NULL ; ConnRemove may already have run on it
			
				invoke ConnRemove, lpCurrentSocketConn

				;invoke shutdown, currentSocket, STOP_SENDS 
				;invoke closesocket, currentSocket

				invoke PostToLog, addr CloseReq

			.endif
;
; FD_ACCEPT occurs when a server receives a request to connect
;		from a client
;
		.elseif wsaAsyncEvent == FD_ACCEPT 
	
			invoke AcceptConn, currentSocket
;
; FD_READ occurs when a data is inbound from the other end
;		of the socket connection
;
		.elseif wsaAsyncEvent == FD_READ 


			invoke RecvData, lpCurrentSocketConn
;
; FD_WRITE occurs when one party of the socket connection 
;		requests to sendData
;
		.elseif wsaAsyncEvent == FD_WRITE 


			invoke SendData ; ignores currentSocket and lpCurrentSocketConn

		.endif

; <<<<<<<<<<<<<<<< >>>>>>>>>>>>>>>>>

	.ELSE
	
		invoke DefWindowProc, hWnd, uMsg, wParam, lParam
		ret
		
	.ENDIF

	xor	eax, eax
	ret

WndProc endp

; -------------------------------------------------------------------------
;		PostToLog PROTO :DWORD
;		Description: Writes output in the main edit window of the
;			module; this is a viewable log; user has option to write
;			contents to a file.
;		Return: Nothing.
; -------------------------------------------------------------------------
PostToLog PROC outputBuffer:DWORD

LOCAL chkLen

	invoke szTrim, outputBuffer
	mov chkLen, eax

	.if eax == 0

		invoke szCopy, chr$( "*****" ), outputBuffer ; self-defense code

	.endif
	
		;	tag on a carriage return
	invoke lstrcat, outputBuffer, addr CrLf

		;	now want to set the selection point at the end
		;		of the existing text in edit buffer, so new
		;		text will be appended, not overwrite what's there
	invoke SendMessage, hwndEdit, EM_SETSEL, MAXLOGTEXT_SIZE, MAXLOGTEXT_SIZE

		;	put the text into the editor window
	invoke SendMessage, hwndEdit, EM_REPLACESEL, NULL, outputBuffer

		;	reset the focus
	invoke SetFocus, hwndEdit

	xor eax, eax
	ret

PostToLog endp

; -------------------------------------------------------------------------
;	ErrorMsg PROTO :DWORD, :DWORD
;		Description: Sends error messages to a message box with 
;			module caption to assist in tracking source of message.
;		Return: Nothing.
; -------------------------------------------------------------------------
ErrorMsg PROC lpMsg:DWORD, lpCap:DWORD

LOCAL hwnd			:HWND

	invoke GetActiveWindow
	mov hwnd, eax

	invoke MessageBox, hwnd, lpMsg, lpCap, MB_OK or MB_ICONWARNING

	xor eax, eax
	ret 
	
ErrorMsg endp

; -------------------------------------------------------------------------

end start
