;
;ServerSupport.asm
;

.386
.model flat, stdcall
option casemap:none

; ------------------------------------------------------------

include Server.inc

; ------------------------------------------------------------

.const

LOC_BUF_SIZE					equ 128

.data

blankStr									db " ", 0

ServerConn							db "Server Conn Mgmt", 0
loginStr									db "Please login.", 0
loginStrSize							equ SIZEOF loginStr ; assume this includes term. 0
invalidClientName					db "Login name invalid", 0
invalidClNameSize					equ SIZEOF invalidClientName

goodbyeFmt							db "Goodbye, %s, have a nice day!", 0
welcomeFmt							db "Welcome to my world, %s!", 0
invalidNameFmt						db "Invalid login name: %s", 0
loggedInFmt							db "Logged in new user: %s", 0
loggedOutFmt							db "Logged out user: %s", 0
forwardDataFmt						db "Forwarding data to: %s", 0
recipNotFndFmt						db "Recipient %s not found - ignored", 0
uploadStartFmt						db "Upload started: Sender: %s, Recipient: %s", 0 
uploadEmptyFmt					db  "Received input from %s without command",0

unkSender								db "*** Sender Unknown ***", 0
wsaBindErr								db "Bind", 0
wsaListenErr							db "Listen", 0
wsaStartErr								db "WSAStartup", 0
wsaSockErr								db "Socket", 0
wsaSyncErr								db "WSAAsyncSelect", 0
connectErr								db "FD_CONNECT", 0
acceptErr									db "FD_ACCEPT", 0
;acceptWelcErr						db "FD_ACCEPT welcome", 0
acceptAttachErr						db "FD_ACCEPT attach", 0
readErr									db "FD_READ", 0
readRecvErr							db "FD_READ recv", 0
writeErr									db "FD_WRITE", 0

lpHeadSocketConn				LPSOCKETCONN 0
lpHeadSendPacket					LPSENDPACKET 0
lpTailSendPacket					LPSENDPACKET 0

.data?

recvBuffer								db BUFFER_SIZE dup (?)
sendBuffer								db BUFFER_SIZE dup (?)
locBuf										db LOC_BUF_SIZE dup (?)
senderName							db CLIENT_NAME_SIZE dup (?)
targClientName						db CLIENT_NAME_SIZE dup (?)
wsadata									WSADATA	<> 

; ------------------------------------------------------------

.code

; ------------------------------------------------------------
;	InitListen PROTO :HWND, :LPSOCKETCONN
;		Description: Initialize Winsock for the server function-
;			ality; create listening socket; sets WSAAsyncSelect
;			for asynchronous operations; binds listening socket;
;			starts server listening for requests to connect.
;		Return: Flag indicating success or failure to initialize.
; ------------------------------------------------------------

InitListen PROC hwnd:HWND, lpListenConn:LPSOCKETCONN 

LOCAL socketListen	:SOCKET
LOCAL sockLstnAddr	:sockaddr_in
LOCAL retVal :DWORD

comment ^
;commented here is a way to use Winsock 2.0
		invoke WSAStartup, REQ_WINSOCK_VER, addr wsadata
		invoke wsprintf, addr buffer, addr highVer, wsadata.wHighVersion 
		invoke MessageBox, hwnd, addr buffer, addr AppName, MB_OK
^

	mov retVal, 0
	
	invoke WSAStartup, 0101h, addr wsadata
	
	.if eax != NULL

		invoke WSAputError,addr wsaStartErr
		xor eax, eax

	.else
	
			; create the listening socket
		invoke socket, AF_INET, SOCK_STREAM, IPPROTO_TCP 

		.if eax == INVALID_SOCKET

			invoke WSAputError,addr wsaSockErr

		.else

			mov socketListen, eax 

				; add asynchronous behavior messages
			invoke WSAAsyncSelect, eax, hwnd, WM_SOCKET, 
							FD_READ+\
							FD_CLOSE+\
							FD_ACCEPT+\
							FD_WRITE
							;don't use FD_CONNECT ; clients only need this

			.if eax == SOCKET_ERROR

				invoke WSAputError,addr wsaSyncErr

			.else

				mov ax, AF_INET
				mov sockLstnAddr.sin_family, ax

				xor eax, eax
				mov sockLstnAddr.sin_addr, eax

				invoke htons, SERVER_PORT
				mov sockLstnAddr.sin_port, ax

				invoke bind, socketListen, addr sockLstnAddr, SIZEOF sockaddr_in		; bind

				.if eax == SOCKET_ERROR

					invoke WSAputError,addr wsaBindErr

				.else

					invoke listen, socketListen, 1									;listen

					.if eax == SOCKET_ERROR

						invoke WSAputError,addr wsaListenErr
	
					.else

						ASSUME ecx:PTR SOCKETCONN
						mov ecx, lpListenConn

						push socketListen
						pop [ecx].hSock

						push sockLstnAddr.sin_family
						pop [ecx].sockRemoteAddr.sin_family

						push sockLstnAddr.sin_addr
						pop [ecx].sockRemoteAddr.sin_addr

						push sockLstnAddr.sin_port
						pop [ecx].sockRemoteAddr.sin_port

						mov retVal, 1

						ASSUME ecx:nothing

					.endif ; listen

				.endif ; bind

			.endif ; WSAAsyncSelect

		.endif ; after socket
		
	.endif ; after WSAStartup

	mov eax, retVal
	ret ; error if any is eax == 0

InitListen endp

; ------------------------------------------------------------
;	AcceptConn PROTO :SOCKET
;		Description: used after wsaAsyncEvent == FD_ACCEPT is
;			reached; establishes a connection from a client by
;			accepting the connect, transferring listen socket charac-
;			teristics, creating a SOCKETCONN data structure for 
;			the connection, and sending a welcome message
;		Return: TRUE if connect okay, otherwise FALSE
; ------------------------------------------------------------

AcceptConn PROC socketListen :SOCKET

LOCAL	remoteAddrSize :DWORD
LOCAL sockConnAddr	:sockaddr_in
LOCAL currentSocket :SOCKET
LOCAL addedConn:LPSOCKETCONN 

	mov eax, SIZEOF sockaddr_in
	mov remoteAddrSize, eax ;needs to send address

	invoke accept, socketListen, addr sockConnAddr, addr remoteAddrSize

	.if eax == INVALID_SOCKET

		invoke WSAputError,addr acceptAttachErr
		mov eax, FALSE

	.else

		mov currentSocket, eax
		invoke ConnAdd, currentSocket, addr sockConnAddr

		mov addedConn, eax

			; send message to client in response to connect
		invoke QueueSendPacket, addedConn, addr loginStr, loginStrSize, NULL, svrCircuit

		invoke PostToLog, chr$( "Accepted client connection" )
		mov eax, TRUE

	.endif

	ret
	
AcceptConn endp	

; ------------------------------------------------------------
;	RecvData PROTO :LPSOCKETCONN
;		Description: Processes incoming data. Parses incoming 
;			buffer for command in the first byte. There are 4
;			commands:
;				'H' == client name ("hello") login
;				'B' == client logging off ("bye")
;				'U' == upload data for relay to a recipient
; 			'W' == request to know who else is logged in
;		Note: only the upload command has data to relay to a 
;			recipient; uploaded data may be in a stream and take 
;			multiple iterations through the recv loop.
;		Note: every command processed will receive feedback in
;			the form of server response, which is a single byte at
;			the head of the response stream (format is handled by the
;			SendData function).	
;		Note: as long as any data is queued to send, whether as a
;			response to the sender or to the recipient, a message is 
;			posted to the main message loop (WndProc) indicating 
;			FD_WRITE, which will therefore be handled asynchronously.
;		Note: the recvBuffer is used as a scratch buffer after its
;			contents have been processed.
;		Return: nothing
; ------------------------------------------------------------

RecvData PROC lpCurSocketConn:LPSOCKETCONN

LOCAL bytesRecvd :DWORD
LOCAL lpTargSocketConn :DWORD
LOCAL stopLooping :DWORD
LOCAL targClientNameLen :DWORD

	mov stopLooping, 0
	mov lpTargSocketConn, 0

	ASSUME ecx:PTR SOCKETCONN

	mov ecx, lpCurSocketConn
	movzx eax, [ecx].sockID
	
	.if eax == CLIENT_NAME_READY

		lea edx, [ecx].clientName
		invoke szCopy, edx, addr senderName

	.endif

	.repeat ; until stopLooping == 1

			; always re-initialize the receive buffer as all of
			;		its previous contents have been processed, if only
			;		temporarily, e.g., into a queue
		invoke memfill, addr recvBuffer, BUFFER_SIZE, 0
		mov ecx, lpCurSocketConn

; <<<<<<<<<<<< RECV --> the main function >>>>>>>>>>>>

		invoke recv, [ecx].hSock, addr recvBuffer, BUFFER_SIZE, 0 ; 

; <<<<<<<<<<<< RECV --> the main function >>>>>>>>>>>>

		.if eax == SOCKET_ERROR

			invoke WSAGetLastError

			.if eax != WSAEWOULDBLOCK

				invoke WSAputError, addr readRecvErr
				
			.else ; was a blocking situation -- so no more data

				.if lpTargSocketConn != 0 

						; if previously receiving an upload for target 
						;		recipient, must now be finished, so send a 
						;		message to sender that data is being forwarded
						;		to the recipient
					invoke wsprintf, addr recvBuffer, addr forwardDataFmt, addr targClientName
					invoke szLen, addr recvBuffer
					invoke QueueSendPacket, lpCurSocketConn, addr recvBuffer, \
																	eax, \
																	NULL, \
																	srvUploadCompl

					invoke PostToLog, addr recvBuffer ; log same message
	
				.endif

			.endif

			mov stopLooping, 1 ; stop receiving upload for any error

		.else ; not a SOCKET_ERROR

			.if eax == 0 ; connection was closed

				mov stopLooping, 1 
				invoke ConnRemove, lpCurSocketConn ; downstream effects???

			.else

				mov bytesRecvd, eax

				mov ecx, offset recvBuffer
				add ecx, eax
				mov BYTE PTR [ecx], 0 ; garbage in there, mask off

				invoke PostToLog, addr recvBuffer

					; it is possible that a send command was received in
					;		a previous iteration through this loop and that
					;		subsequent reads may occur; this data will be part
					;		of that stream and therefore will not carry a command;
					;		it merely needs to be queued up and forwarded to
					;		the recipient to follow the previous queued batch(es)

				.if lpTargSocketConn != 0 ; already loading for target recipient

					invoke QueueSendPacket, lpTargSocketConn, addr recvBuffer, \
																	bytesRecvd, \
																	NULL, \
																	srvDownload

				.else ; must be a new send upload

;
;parse the input to get the command in 1st byte
;
	
					mov edx, OFFSET recvBuffer
					push edx
					add edx, bytesRecvd
					mov BYTE PTR [edx], 0 ;mask extra CRLF at end of buffer
					pop edx
					mov al, BYTE PTR [edx]  ;command character is in 1st byte
;
;	Client module is logging in with a name string; name must
;		meet size restrictions and must not have embedded blanks
;	

					.if al == cmdHello  ; <<<<<<<<<<< 'H' command

						invoke memfill, addr senderName,  CLIENT_NAME_SIZE, 0

							; note: client prog won't let user log in more than once
							; login name begins in byte after command
							; only data after cmd in recvBuffer will be login name
						mov edx, OFFSET recvBuffer
						add edx, 1 
						mov eax, bytesRecvd
						sub eax, 1
						invoke MemCopy, edx, addr senderName, eax

						invoke ClientNameSet, addr senderName, lpCurSocketConn

						.if eax == 0 ; ClientNameSet rejected invalid name

							invoke QueueSendPacket, lpCurSocketConn, addr invalidClientName, \
																			invalidClNameSize, \
																			NULL, \
																			srvError

							invoke wsprintf, addr recvBuffer, addr invalidNameFmt, addr senderName
							invoke PostToLog, addr recvBuffer

						.else

							invoke wsprintf, addr recvBuffer, addr welcomeFmt, addr senderName
							invoke szLen, addr recvBuffer

							invoke QueueSendPacket, lpCurSocketConn, addr recvBuffer, \
																			eax, \
																			NULL, \
																			srvLogCompl

							invoke wsprintf, addr recvBuffer, addr loggedInFmt, addr senderName
							invoke PostToLog, addr recvBuffer

						.endif
	
						mov stopLooping, 1
;
;	Client module is logging out; server will shut down the socket
;		gracefully after acknowledging client's action
;
					.elseif al == cmdGoodbye ; <<<<<<<<<<< 'B' command


						invoke wsprintf, addr recvBuffer, addr goodbyeFmt, addr senderName
						invoke szLen, addr recvBuffer


						invoke QueueSendPacket, lpCurSocketConn, addr recvBuffer, \
																		eax, \
																		NULL, \
																		srvTerminated

						invoke wsprintf, addr recvBuffer, addr loggedOutFmt, addr senderName
						invoke PostToLog, addr recvBuffer

						mov ecx, lpCurSocketConn
						mov [ecx].sockState, STAT_CLOSE_PENDING  ; SendData uses this flag

						mov stopLooping, 1

;
;	Client module has requested a "Who" list of names of logged
;		in clients; when data is uploaded, it requires a valid
;		client Id to forward message to.
;
					.elseif al == cmdWho ; <<<<<<<<<<< 'W' command

						invoke memfill, addr recvBuffer, BUFFER_SIZE, 0
						invoke ClientNameList, addr recvBuffer

						invoke QueueSendPacket, lpCurSocketConn, addr recvBuffer, \
																		eax, \
																		NULL, \
																		svrInfo

						mov stopLooping, 1
;
;	Client module is uploading a block of data; note that the
;		format of an upload is "UclientnameXmessagetexthere",
;		where 'U' is the upload command; followed by recipient
;		client name which has length requirements and may not have
;		embedded blanks -- and must be logged in; 'X' is the bounding 
;		blank; and then comes the text of the message -- which has a 
;		maximum length here of 07FFFh (system default, not overridden).
;
					.elseif al == cmdUpload ; <<<<<<<<<<< 'U' command

;	does not stop the looping unless the recipient is invalid;
;	start by parsing out recipient, begins in 2nd char of
;		input in recvBuffer and is terminated by blank space

						invoke memfill, addr locBuf, LOC_BUF_SIZE, 0 ; clear out scratch buffer

						mov edx, OFFSET recvBuffer
						add edx, 1 ; recipient name begins next byte after command

						invoke MemCopy, edx, addr locBuf, CLIENT_NAME_SIZE

						invoke szTrim, addr locBuf

						.if eax == 0 ;name field was blanks

							mov stopLooping, 1 ; client software should prevent this

						.else

								;blank space terminates name, so rep blank with zero
							invoke InString, 1, addr locBuf, addr blankStr

							.if eax > 0

								sub eax, 1

								mov targClientNameLen, eax

								lea edx, locBuf
								add edx, eax
								mov BYTE PTR [edx], 0 ; put zero terminator

							.endif
								; parse out the recipient name
							invoke szCopy, addr locBuf, addr targClientName

							invoke ClientNameFind, addr targClientName ; see if logged in

							.if eax == 0 ; not logged in
							
								mov stopLooping, 1 ; recipient name requested is invalid
								
								invoke wsprintf, addr recvBuffer, addr recipNotFndFmt, addr targClientName
								invoke szLen, addr recvBuffer
								invoke QueueSendPacket, lpCurSocketConn, addr recvBuffer, \
																				eax, \
																				NULL, \
																				srvError

								invoke PostToLog, addr recvBuffer ; log same message
	
							.else	; have a match for target name -- got its SOCKETCONN

								mov lpTargSocketConn, eax ; SOCKETCONN for recipient

								invoke wsprintf, addr locBuf, addr uploadStartFmt, \
																addr senderName, addr targClientName
								invoke PostToLog, addr locBuf

								mov eax, targClientNameLen
								add eax, 2 ; 1 byte for command + 1 byte for space

								mov edx, OFFSET recvBuffer
								add edx, eax ; offset 1st byte after target name & space

								sub bytesRecvd, eax

									; now can queue up data starting in recvBuffer
									;		after command+recipient name+blank space
								invoke QueueSendPacket, lpTargSocketConn, edx, \
																				bytesRecvd, \
																				addr senderName, \
																				srvDownload

							.endif

						.endif
						
					.else ; recv buffer arrived w/o a command 

						invoke wsprintf, addr locBuf, addr uploadEmptyFmt, \
														addr senderName
						invoke PostToLog, locBuf
	
						mov stopLooping, 1

					.endif

				.endif

			.endif

		.endif

	.until stopLooping == 1

	ASSUME ecx:nothing
	
	xor eax,eax ;nothing to return
	ret
	
RecvData endp	

; ------------------------------------------------------------
;	SendData PROTO 
;		Description: Receives no input as it merely has to work
;			its way through the linked list queue of SENDPACKETs 
;			and forward those to their respective recipients; works 
;			in a loop as it dequeues the items in the queue and
;			forwards them; stops the loop when either the queue is
;			emptied or an attempt to send results in any error;
;			most common will be an error equal to WSAEWOULDBLOCK
;		Return: nothing
; ------------------------------------------------------------

SendData PROC

LOCAL stopLooping 				:DWORD
LOCAL lpBuffer 						:DWORD
LOCAL bytesSent 					:DWORD
LOCAL sendCnt 						:DWORD
LOCAL prevBytesSent				:DWORD
LOCAL bufferSize 					:DWORD
LOCAL sendSock						:SOCKET
LOCAL	lpCurSendPacket 		:LPSENDPACKET
LOCAL lpCurSocketConn		:LPSOCKETCONN

	push esi
	
	mov stopLooping, 0
	mov lpCurSendPacket, NULL
	
	ASSUME ecx:PTR SENDPACKET

	.repeat ; until stopLooping == 1
	
		invoke memfill, addr sendBuffer,  BUFFER_SIZE, 0

		.if lpCurSendPacket == NULL ; see if 1st time through loop
		
			invoke DequeueSendPacket ; get next enqueued SENDPACKET 
		
		.else ; still working on previous SENDPACKET

			mov eax, lpCurSendPacket

		.endif

		.if eax == NULL ; queue is empty, so stop sending

			mov stopLooping, 1

		.else ; have data to send to client

			mov lpCurSendPacket, eax
			mov ecx, eax ; address of current SENDPACKET

			push [ecx].lpSockConn
			pop lpCurSocketConn

			push [ecx].bytesSent
			pop edx
			mov prevBytesSent, edx

			push [ecx].bufferSize
			pop eax
			mov bufferSize, eax
			sub eax, edx ; eax now remainder to send
			mov sendCnt, eax

			mov esi, [ecx].lpSendBuffer ; get addr of send buffer
			add esi, edx ; advance ptr to next byte to send

			invoke MemCopy, esi, addr sendBuffer, sendCnt

			mov edx, lpCurSocketConn

			ASSUME edx:PTR SOCKETCONN

			push [edx].hSock
			pop sendSock

; <<<<<<<<<<<< SEND --> the main function >>>>>>>>>>>>

			invoke send, sendSock, addr sendBuffer, sendCnt, 0

; <<<<<<<<<<<< SEND --> the main function >>>>>>>>>>>>

			.if eax == SOCKET_ERROR

				invoke WSAGetLastError

				.if eax != WSAEWOULDBLOCK

					invoke WSAputError, addr readRecvErr
					
						; assume that the recipient connection has been lost
						;		so we need to remove its SENDPACKETs;
						;	the ConnRemove process will physically (yet grace-
						;		fully) remove the SOCKETCONN and all queued up 
						;		SENDPACKETs.

					invoke ConnRemove, lpCurSocketConn

				.endif

					; set up another go-around for sends
				invoke GetActiveWindow
				mov ecx, FD_WRITE ; high word == 0
				mov edx, lpCurSocketConn ; FD_WRITE ignores SOCKETCONN & SOCKET
				invoke PostMessage, eax, WM_SOCKET, [edx].hSock, ecx

				mov stopLooping, 1 ; stop for all error conditions

			.else ; at least something was sent

				mov bytesSent, eax ; count bytes actually sent
				add eax, prevBytesSent

				mov ecx, lpCurSendPacket
				push eax
				pop [ecx].bytesSent ; bytesSent checked by DequeueSendPacket
				
				.if eax == bufferSize

					mov lpCurSendPacket, NULL ; through with this SENDPACKET; dequeue another

					mov edx, lpCurSocketConn
					movzx eax, [edx].sockState

						; special case where client sent "Goodbye" acknowledgement
					.if eax == STAT_CLOSE_PENDING 

							; SOCKETCONN can be removed since final message sent
						invoke ConnRemove, edx
						mov stopLooping, 1

					.endif

				.endif

			.endif

			ASSUME edx:nothing

		.endif

	.until stopLooping == 1

	ASSUME ecx:nothing

	pop esi

	xor eax,eax
	ret
	
SendData endp	

; ------------------------------------------------------------
;	QueueSendPacket	PROTO :LPSOCKETCONN, :DWORD, :DWORD, :BYTE
;		Description: Receives a buffer of data that is pending
;			transmission to a recipient; allocates memory for
;			holding SENDPACKET buffer contents; the srvrFlag is
;			placed at the beginning of the SENDPACKET buffer -- it
;			is the signal the recipient uses to interpret the con-
;			tents of the transmission.
;		Note: the sendPacketQueue is a FIFO linked list queue.
;		Note: lenBuf is assumed to exclude terminating 0.
;		Note: if Alloc fails to allocate storage, the enqueueing 
;			is bypassed and data is not transmitted.
;		Note: if the send packet is a message going to a target
;			client, it will have the same form as the uploaded 
;			message, namely: "DsendernameXmessagetext", where 'D'
;			is a flag to client that it's a download message, 
;			sendername is the sender's login ID, 'X' is a blank
;			separator, and messagetext is the message text. If not
;			a download, the sender name argument is NULL.
;		Return: nothing
; ------------------------------------------------------------

QueueSendPacket PROC lpCurSocketConn:LPSOCKETCONN, \
												lpBufIn:DWORD, \
												lenBuf:DWORD, \
												lpSenderName:DWORD, \
												srvrFlag:BYTE

LOCAL	lpBuffer :DWORD
LOCAL	dataSize :DWORD
LOCAL lenSenderName :DWORD
LOCAL lpCurSendPacket :LPSOCKETCONN

	mov lenSenderName, 0
	
	mov eax, lenBuf
	add eax, 2 ; add one temporarily for 0 byte after
	mov dataSize, eax

	.if lpSenderName != NULL ; compute size requirements for sender name

		invoke szLen, lpSenderName
		add eax, 1 ; add 1 for trailing blank
		mov lenSenderName, eax
		add dataSize, eax

	.endif

	mov eax, dataSize
	shr eax, 2
	add eax, 1
	shl eax, 2 ; make evenly divisible by 4
	mov dataSize, eax
	
	invoke Alloc, dataSize ; alloc space for holding data while enqueued
	
	.if eax != NULL ; ignores case of no alloc -- data is dropped

		mov lpBuffer, eax
		
		invoke memfill, eax, dataSize, 0

		sub dataSize, 1 ; see above
		mov edx, lpBuffer
		
		.if srvrFlag != srvNULL ; never actually used (yet)

			mov al, srvrFlag
			mov BYTE PTR [edx], al ; put the flag at beginning of buffer
	
			add edx, 1 ; addr for input buffer is 1 after srvrFlag command

			.if lenSenderName != 0

				push edx
				invoke szCopy, lpSenderName, edx
				pop edx
				sub lenSenderName, 1 ; remove count for blank
				add edx, lenSenderName
				mov BYTE PTR [edx],BLANK_CHAR
				add edx, 1 ; to account for blank after sender name

			.endif

		.endif

		invoke MemCopy, lpBufIn, edx, lenBuf ; load buffer with queued data

			; now allocate SENDPACKET to hold the control information
		mov lpCurSendPacket,  halloc( SENDPACKET_SIZE )

		.if lpCurSendPacket == NULL 

			invoke Free, lpBuffer ; failed, don't create SENDPACKET, don't send

		.else ; everything allocated as needed
	
			ASSUME ecx:PTR SENDPACKET ; load SENDPACKET control struct
		
			invoke memfill, lpCurSendPacket, SENDPACKET_SIZE, 0
	
			.if lpHeadSendPacket == 0 ; 1st entry in linked list queue?

				mov ecx, lpCurSendPacket
				mov lpHeadSendPacket, ecx ; this is first, so put into head ptr

			.else ; add new SENDPACKET at end of linked list

				mov ecx, lpTailSendPacket 
				push lpCurSendPacket
				pop [ecx].nextSendPacket ; tail's next points to new SENDPACKET
				mov ecx, lpCurSendPacket

			.endif

			push lpCurSocketConn		; load up SENDPACKET data
			pop [ecx].lpSockConn

			push lpBuffer
			pop [ecx].lpSendBuffer
	
			push dataSize
			pop [ecx].bufferSize
	
			mov [ecx].bytesSent, 0

			push ecx
			pop lpTailSendPacket ; tail now contains ptr to new SENDPACKET

			ASSUME ecx:PTR SOCKETCONN

				; if any items are queued to send to recipients, the 
				;		main message loop needs to be apprised of that fact;
				;		send the WM_SOCKET/FD_WRITE msg and let the SendData
				;		function sort it out; after all, all data is queued
				;		up and can be sent in an asychronous manner

			invoke GetActiveWindow
			mov edx, FD_WRITE ; high word == 0
			mov ecx, lpCurSocketConn
			invoke PostMessage, eax, WM_SOCKET, [ecx].hSock, edx

			ASSUME ecx:nothing

		.endif

	.endif

	xor eax,eax ; returns nothing -- failure with Alloc ignored!
	ret
	
QueueSendPacket endp	

; ------------------------------------------------------------
;	DequeueSendPacket PROTO
;		Description: Extracts next enqueued SENDPACKET for trans-
;			mission to a recipient; works through a linked list.
;		Note: Removal discipline is FIFO (standard queue).
;		Note: Before extracting next SENDPACKET, checks to see if
;			current SENDPACKET has been completely sent; if not complete,
;			returns current SENDPACKET; otherwise, deletes the Alloc'd
;			storage and reinitializes that SENDPACKET, and then
;			advances pointer to next entry in the queue.
;		Return: SENDPACKET with data to transmit, NULL if nothing
;			left to send.
; ------------------------------------------------------------

DequeueSendPacket PROC 

LOCAL	lpCurSendPacket :LPSENDPACKET

	push lpHeadSendPacket
	pop lpCurSendPacket
	
	.if lpCurSendPacket != NULL ; otherwise queue is empty

		ASSUME ecx:PTR SENDPACKET

		mov ecx, lpCurSendPacket

			;now determine if all data in SENDPACKET already sent
		push [ecx].bufferSize
		pop eax
	
		push [ecx].bytesSent
		pop edx

		sub eax, edx

		.if eax == 0 ; all data sent from current SENDPACKET, deallocate

			push ecx ;**
			push [ecx].lpSendBuffer
			pop eax
			invoke Free, eax ; release memory of buffered data

			pop ecx ;**
			push [ecx].nextSendPacket
			hfree( ecx ) ; release memory of the SENDPACKET
			pop ecx ; point to next SENDPACKET if not NULL

			mov lpCurSendPacket, ecx

			;sub cntQueuedPackets, 1

		.endif

		push ecx
		pop lpHeadSendPacket

		ASSUME ecx:nothing
	
	.endif

	mov eax, lpCurSendPacket
	ret
	
DequeueSendPacket endp	

; ------------------------------------------------------------
;	ClearSendQueue	PROTO :LPSOCKETCONN
;		Description: travels through entire sendqueue to see if
;			there are any queued SENDPACKETs for lpRmvConn; deletes 
;			them if so; if lpRmvConn is NULL, deletes all SENDPACKETs.
;		Returns: nothing
; ------------------------------------------------------------

ClearSendQueue PROC lpRmvConn :LPSOCKETCONN 

LOCAL	lpCurSendPacket :LPSENDPACKET
LOCAL	lpPrevSendPacket :LPSENDPACKET
LOCAL rmvFlag :DWORD

	push lpHeadSendPacket
	pop lpCurSendPacket
	
	mov lpPrevSendPacket, NULL 

	.if lpCurSendPacket != NULL ; otherwise queue is empty

		ASSUME ecx:PTR SENDPACKET

		.repeat ; .until lpCurSendPacket == NULL

			mov ecx, lpCurSendPacket
			mov rmvFlag, 0

			.if lpRmvConn == NULL 

				mov rmvFlag, 1 ; remove all SENDPACKETs

			.else ; removing SENDPACKETs for specific SOCKETCONN 

				mov eax, [ecx].lpSockConn

				.if eax == lpRmvConn ; check for match

					mov rmvFlag, 1 ; match

					push ecx ;***

						; before removing SENDPACKET, de-link it
					push [ecx].nextSendPacket

					.if lpPrevSendPacket == NULL ; cur SENDPACKET is first in queue
					
						pop lpHeadSendPacket ; now points to second in queue

					.else

						mov ecx, lpPrevSendPacket 
						pop [ecx].nextSendPacket

					.endif

					pop ecx ;***

				.else

					mov lpPrevSendPacket, ecx ; this one not removed, but flagged

				.endif

			.endif

			push [ecx].nextSendPacket
			pop lpCurSendPacket

			.if rmvFlag == 1

				push ecx ;***
				push [ecx].lpSendBuffer
				pop eax 
				invoke Free, eax ; release buffer memory

				pop ecx ;***
				hfree( ecx ) ; release SENDPACKET memory

			.endif

		.until lpCurSendPacket == NULL

		ASSUME ecx:nothing

	.endif
	
	xor eax, eax
	ret

ClearSendQueue endp

; ------------------------------------------------------------
;	ConnAdd PROTO :SOCKET, :DWORD
;		Description: finds the end of the linear linked list of 
;			SOCKETCONN's for the connected clients, and adds the new 
;			socket into its own SOCKETCONN.
;		Returns: ptr to new connectee's SOCKETCONN
; ------------------------------------------------------------

ConnAdd PROC sock:SOCKET, lpSockAddr_In:DWORD

LOCAL	lpConnData :DWORD
LOCAL curSockConn :LPSOCKETCONN
LOCAL nextSockConn :LPSOCKETCONN

	mov lpConnData,  halloc( SOCKETCONN_SIZE )

	.if lpConnData != 0 
	
		ASSUME ecx:PTR SOCKETCONN
		
		invoke memfill, lpConnData, SOCKETCONN_SIZE, 0
	
		.if lpHeadSocketConn == 0

			mov ecx, lpConnData
			mov lpHeadSocketConn, ecx ; this is first, so put into head ptr

		.else ; add new SOCKETCONN at end of linked list

			mov ecx, lpHeadSocketConn

				; search for end of linked list, node == 0
			.repeat ; .until nextSockConn == 0

				push [ecx].nextSocketConn
				pop nextSockConn 

				.if nextSockConn != 0
	
					mov ecx, nextSockConn ; ecx points to next link
	
				.endif

			.until nextSockConn == 0

				; ecx now points to last link whose 'next' gets new link
			push lpConnData
			pop [ecx].nextSocketConn

			mov ecx, lpConnData ; now ecx points to new SOCKETCONN

		.endif

		push sock ; load data into the new SOCKETCONN
		pop [ecx].hSock

		mov [ecx].sockID, CLIENT_NAME_NULL ; name entered with login
		mov [ecx].sockState, STAT_READ_READY ; status

		lea edx, [ecx].sockRemoteAddr
		invoke MemCopy, lpSockAddr_In, edx, SOCKADDR_SIZE

		mov eax, lpConnData ; return ptr to new connectee SOCKETCONN

		ASSUME ecx:nothing
	
	.else
	
		invoke ErrorMsg,  chr$( " Allocating connectee space" ), addr ServerConn
		xor eax, eax
		
	.endif
	
	ret

ConnAdd endp

;	------------------------------------------------------------
;	ConnFind PROTO :SOCKET
;		Description: locates previously allocated SOCKETCONN 
;			structure data for a newly arrived client input;
;			search is made through the linear linked list of
;			SOCKETCONN's.
;		Return: ptr to SOCKETCONN data if found, else 0.
; ------------------------------------------------------------

ConnFind PROC hSock:SOCKET

	mov edx, hSock
	mov ecx, lpHeadSocketConn
	
	ASSUME ecx:PTR SOCKETCONN
	
		; search for socket in linked list
	.while ecx != 0

		cmp [ecx].hSock, edx
		jne @F
	
		.break
	
@@:
		push [ecx].nextSocketConn
		pop ecx ; ecx points to next link
		
	.endw 
	
	ASSUME ecx:nothing

	mov eax, ecx ; return ptr to SOCKETCONN data if found, else 0
	ret
	
ConnFind endp

; ------------------------------------------------------------
;ConnRemove PROTO :LPSOCKETCONN
;		Description: remove SOCKETCONN structure data for a client
;			that is closing its connection.
;		Return: flag indicating success of removal.
; ------------------------------------------------------------

ConnRemove PROC lpRmvSocketConn:LPSOCKETCONN

LOCAL fnd :DWORD
LOCAL sock :SOCKET

	mov fnd, 0

	mov edx, lpRmvSocketConn ; SOCKETCONN to remove
	mov ecx, lpHeadSocketConn ; first SOCKETCONN in linked list
	
	ASSUME ecx:PTR SOCKETCONN
	
	.if edx == ecx ; test if it's the head node in the linked chain

		push [ecx].nextSocketConn
		pop lpHeadSocketConn ; set head node ptr to second in chain

		mov fnd, 1 ; flag so we can free this memory later
	
	.else

		push [ecx].nextSocketConn ; traverse chain to second node
		pop ecx ; ecx now points to second node
		
			; begin searching for the removable SOCKETCONN in linked list
		.while ecx != 0
		
			.if ecx == edx
	
				mov fnd, 1 ; ecx matched the removable node
				.break ; stop searching now
	
			.else

				push [ecx].nextSocketConn
				pop ecx ; ecx points to next link

			.endif

		.endw

	.endif
	
	.if fnd == 1

		push ecx

		push [ecx].hSock 
		pop sock
		invoke ClearRecvBufs, sock
		invoke shutdown, sock, STOP_SENDS ; gracefully disconnect
		invoke closesocket, sock

		invoke ClearSendQueue, lpRmvSocketConn

		pop ecx
		hfree( ecx )
	
	.endif
	
	ASSUME ecx:nothing
	
	mov eax, fnd ; return flag whether SOCKETCONN was removed
	ret
	
ConnRemove endp

; ------------------------------------------------------------
; GracefulDisconAll	PROTO
;		Description: disconnect all open connections and release
;			the memory; used for closing down.
;		Return: nothing.
; ------------------------------------------------------------

GracefulDisconAll PROC 

LOCAL sock :SOCKET

	mov ecx, lpHeadSocketConn ; first SOCKETCONN in linked list
	
	ASSUME ecx:PTR SOCKETCONN
	
	.while ecx != 0

		push ecx

		push [ecx].hSock
		pop sock
		invoke ClearRecvBufs, sock
		invoke shutdown, sock, STOP_SENDS
		invoke closesocket, sock

		pop ecx
		
		push [ecx].nextSocketConn ; traverse chain
		hfree( ecx )
		pop ecx ; ecx now points to next node
	
	.endw
	
	ASSUME ecx:nothing

	xor eax, eax
	ret
	
GracefulDisconAll endp

; ------------------------------------------------------------
; ClearRecvBufs	PROTO :SOCKET
;		Description: execute recv on socket until all buffers have
;			been cleared; documentation indicates this is needed
;			even if no FD_READ was received -- to be on safe side?
;		Return: nothing.
; ------------------------------------------------------------

ClearRecvBufs PROC sock:SOCKET 

LOCAL stopLoop :DWORD

	mov stopLoop, 0

	.repeat ; .until stopLoop == 1

		invoke recv, sock, addr recvBuffer, BUFFER_SIZE, 0 ; 

		.if eax == SOCKET_ERROR || eax == 0

			mov stopLoop, 1

		.endif

	.until stopLoop == 1

	xor eax, eax
	ret

ClearRecvBufs endp

; ------------------------------------------------------------
;ClientNameSet	PROTO :DWORD, :LPSOCKETCONN
;		Description: sets the name string into the current connec-
;			tion's SOCKETCONN data structure name area; name will 
;			have blanks removed and will be set to max length of
;			CLIENTNAME_SIZE-1; if name is a duplicate, it's rejected
;			note: make the assumption that the input name string is
;			sitting in a buffer that is at least CLIENTNAME_SIZE
;			bytes long.
;		Return: 0 if name is invalid, len if valid.
; ------------------------------------------------------------

ClientNameSet PROC lpNameString:DWORD, lpCurConn:LPSOCKETCONN

LOCAL sLen :DWORD

	invoke szTrim, lpNameString ; trim blanks/tabs on ends (trims in place)
	
	mov eax, rv( szLen, lpNameString ) ; anything left after trim?
	mov sLen, eax ; save for later
	
	.if eax > 0
	
			; check for blanks within name string
		invoke InString, 1, lpNameString, chr$( " " )
		
		.if eax != 0 ; embedded blanks found if not zero
	
			mov eax, 0 ; invalid
	
		.else
	
				; check for sLen > max length
			.if sLen > CLIENT_NAME_SIZE-1
	
				mov eax, 0 ; too long, invalid
	
			.else

				invoke ClientNameFind, lpNameString
	
				.if eax == 0 ; zero means no duplicate name
	
					mov eax, sLen

				.else
	
					mov eax, 0 ; duplicate name not valid
	
				.endif
	
			.endif
	
		.endif

	.endif

	.if eax > 0 ; name is valid: good len, no spaces, no duplicates

			; prepare to load name into current SOCKETCONN
		ASSUME ecx:PTR SOCKETCONN
	
		mov ecx, lpCurConn 

		mov [ecx]. sockID, CLIENT_NAME_READY
		mov [ecx].sockState, STAT_READ_READY
		lea edx, [ecx].clientName

		push edx
		invoke memfill, edx, CLIENT_NAME_SIZE, 0
		pop edx

		invoke MemCopy, lpNameString, edx, sLen

		mov eax, sLen

		ASSUME ecx:nothing

	.endif
	
	ret ; eax contains length of string moved into clientName area
	
ClientNameSet endp

; ------------------------------------------------------------
;ClientNameFind	PROTO :DWORD
;		Description: find the SOCKETCONN data structure containing
;			the requested client login name string.
;		Return: ptr to SOCKETCONN data structure if found, else 0.
; ------------------------------------------------------------

ClientNameFind PROC lpNameString:DWORD

	mov ecx, lpHeadSocketConn
	
	ASSUME ecx:PTR SOCKETCONN
	
		; search for name in linked list
	.while ecx != NULL

		lea edx, [ecx].clientName

		push ecx
		invoke Cmpi, edx, lpNameString
		pop ecx 
		
		.if eax == 0

			.break ; match found -- ecx has addr of located SOCKETCONN
	
		.else

			push [ecx].nextSocketConn
			pop ecx ; ecx points to next link or to NULL
	
		.endif
		
	.endw 
	
	ASSUME ecx:nothing

	mov eax, ecx ; return ptr to SOCKETCONN data if found, else 0
	ret
	
ClientNameFind endp

; ------------------------------------------------------------
;	ClientNameList	PROTO :DWORD
;		Description: runs through the linked list of SOCKETCONN
;			data structures and extracts client names, creating a
;			space delimited string of all names.
;		Note: assumes the buffer is big enough; also assumes that
;			the buffer has been zero filled.
;		Return: length of data in name list buffer.
; ------------------------------------------------------------

ClientNameList PROC lpNameBuffer:DWORD

	mov ecx, lpHeadSocketConn
	mov eax, 0
	
	ASSUME ecx:PTR SOCKETCONN
	
		; run through linked list and grab client names
	.while ecx != 0

		lea edx, [ecx].clientName

			; lead eax with next append position, i.e., length
		push ecx
		mov eax, rv( szappend, lpNameBuffer, edx, eax ) 
		pop ecx

		push [ecx].nextSocketConn
		pop ecx ; ecx points to next link
	
		.if ecx != 0 ; append blank space if not last name string
		
			push ecx
			mov eax, rv( szappend, lpNameBuffer, chr$( " " ), eax )
			pop ecx

		.endif
	
	.endw 
	
	ASSUME ecx:nothing

		; eax contains length of buffer
	ret
	
ClientNameList endp

; ------------------------------------------------------------

END